import click
import logging

from deltalake_tools.core.core import (
    delta_compact,
    delta_vacuum,
    delta_create_checkpoint,
    table_version as delta_table_version
)
from deltalake_tools.models.models import (
    TableType,
    S3ClientDetails,
    S3Scheme,
    S3KeyPairWrite,
    VirtualAddressingStyle
)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version='0.1.2')
def cli():
    pass


@cli.command()
@click.argument('delta-table-path')
def compact(delta_table_path: str):
    result = delta_compact(delta_table_path)

    if result.is_err():
        print(result.unwrap_err())
    else:
        print(result.unwrap())


@cli.command()
@click.argument('delta-table-path')
@click.option('--retention-hours', default=168, type=int)
@click.option('--disable-retention-duration', is_flag=True)
@click.option('--force', is_flag=True)
def vacuum(delta_table_path: str,
        retention_hours: int,
        disable_retention_duration: bool = False,
        force: bool = False):

    result = delta_vacuum(
                delta_table_path,
                retention_hours=retention_hours,
                enforce_retention_duration=not disable_retention_duration,
                dry_run=not force,
    )

    if result.is_err():
        print(result.unwrap_err())
    else:
        print(result.unwrap())

@cli.command()
@click.argument('delta-table-path')
def create_checkpoint(delta_table_path: str):
    result = delta_create_checkpoint(delta_table_path)

    if result.is_err():
        print(result.unwrap_err())
    else:
        print(result.unwrap())

@cli.command()
@click.argument('delta-table-path')
@click.option('--bucket', required=False, type=str)
@click.option('--access-key-id', required=False, type=str)
@click.option('--secret-access-key', required=False, type=str)
@click.option('--region', required=False, type=str)
@click.option('--endpoint-host', required=False, type=str)
@click.option('--port', type=int, required=False)
@click.option('--scheme', type=click.Choice(['http', 'https']), required=False)
@click.option('--allow-unsafe-https', is_flag=True)
@click.option('--path-addressing-style', is_flag=True)
def table_version(delta_table_path: str,
        *,
        bucket: str,
        access_key_id: str,
        secret_access_key: str,
        region: str = "us-east-1",
        endpoint_host: str,
        port: int = 443,
        scheme: str = "https",
        allow_unsafe_https: bool = False,
        path_addressing_style: bool = VirtualAddressingStyle.Path):
    client_details: S3ClientDetails = None
    # logger.error("in the table version command")
    table_type = check_delta_table_type(delta_table_path)
    if table_type == TableType.S3:
        if bucket is None:
            raise click.BadParameter(
                "Bucket must be provided, along with storage options (see -h), when using S3 table type."
            )  
        # logger.error(f"Bucket: {bucket}")
        client_details_result = S3ClientDetails.default()
        if client_details_result.is_err():
            click.echo(client_details_result.unwrap_err())
        
        client_details = client_details_result.unwrap()

        if bucket is not None:
            client_details.bucket = bucket
        
        if access_key_id is not None and secret_access_key is not None:
            client_details.hmac_keys = S3KeyPairWrite(
                            access_key_id = access_key_id,
                            secret_access_key = secret_access_key)

        if region is not None:
            client_details.region = region
        
        if endpoint_host is not None:
            client_details.endpoint_host = endpoint_host
        
        if port is not None:
            client_details.port = port
        
        if scheme is not None:
            client_details.scheme = S3Scheme(scheme)
        
        if allow_unsafe_https:
            client_details.allow_unsafe_https = True
        
        if path_addressing_style:
            client_details.virtual_addressing_style = VirtualAddressingStyle.Path


    result = delta_table_version(delta_table_path, client_details=client_details)

    if result.is_err():
        print(result.unwrap_err())
    else:
        print(result.unwrap())


def check_delta_table_type(delta_table_path: str) -> TableType:
    if delta_table_path.startswith("s3://"):
        return TableType.S3
    else:
        return TableType.Local


 
from pydantic import BaseModel
from enum import Enum
from typing import Optional

from deltalake_tools.result import Result, Ok, Err

class S3Scheme(Enum):
    Http: str = "http"
    Https: str = "https"

class VirtualAddressingStyle(Enum):
    Virtual: str = "true"
    Path: str = "false"

class S3KeyPairWrite(BaseModel):
    access_key_id: Optional[str]
    secret_access_key: Optional[str]
 
class S3ClientDetails(BaseModel):
    endpoint_host: str = None
    region: str = None
    port: int = None
    virtual_addressing_style: Optional[VirtualAddressingStyle] = VirtualAddressingStyle.Virtual
    bucket: Optional[str] = None
    hmac_keys: Optional[S3KeyPairWrite] = None
    scheme: Optional[S3Scheme] = None
    allow_unsafe_https: bool = False

    @staticmethod
    def default() -> Result['S3ClientDetails', str]:
        return Ok(S3ClientDetails(
            endpoint_host="s3.amazonaws.com",
            region="us-east-1",
            port=443,
            addressing_style=VirtualAddressingStyle.Virtual,
            bucket=None,
            hmac_keys=None,
            scheme=S3Scheme.Https,
            allow_unsafe_https=False,
            )
        )

    def endpoint_url(self) -> Result[str, str]:
        match self.virtual_addressing_style:
            case VirtualAddressingStyle.Virtual:
                return Ok(f"{self.scheme.value}://{self.bucket}.{self.endpoint_host}:{self.port}")
            case VirtualAddressingStyle.Path:
                return Ok(f"{self.scheme.value}://{self.endpoint_host}:{self.port}")

    def to_s3_config(self) -> Result[dict[str, str], str]:


        return Ok({
            "aws_virtual_hosted_style_request": self.virtual_addressing_style.value,
            "AWS_REGION": self.region,
            "AWS_ACCESS_KEY_ID": self.hmac_keys.access_key_id,
            "AWS_SECRET_ACCESS_KEY": self.hmac_keys.secret_access_key,
            "endpoint": self.endpoint_url().unwrap(),
            "AWS_S3_ALLOW_UNSAFE_RENAME": "true",
            "AWS_ALLOW_HTTP": "true" if self.allow_unsafe_https else "false",
        })

class ClientDetails(BaseModel):
    delta_path: str
    s3client_details: Optional[S3ClientDetails] = None

class TableType(Enum):
    S3: str = "S3"
    Local: str = "Local"


from click.testing import CliRunner
from deltalake_tools.cli.cli import (
    vacuum,
    compact,
    create_checkpoint,
    table_version
)
from deltalake import DeltaTable, write_deltalake
import pandas as pd
import pytest
import logging

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

@pytest.fixture(scope="session")
def cli_delta_table_with_data(cli_delta_table_path):
    dt = DeltaTable(cli_delta_table_path)
    # logger.error(f"{cli_delta_table_path=}")

    assert dt.version() == 0

    for i in range(10):
        data = pd.DataFrame({
            "id": [i+2, i+3],
            "name": [f"{str(i)}_name", f"{str(i)}_other_name"]
        })

        write_deltalake(cli_delta_table_path, data, mode="append")

    dt_result = DeltaTable(cli_delta_table_path)

    assert dt_result.version() == 10
    assert len(dt_result.to_pandas()) == 22

    assert len(dt_result.files()) == 11

    # shutil.copytree(cli_delta_table_path, "/tmp/delta_table_backup")

    yield cli_delta_table_path


def test_compact(cli_delta_table_with_data):
    runner = CliRunner()
    result = runner.invoke(compact, [cli_delta_table_with_data])
    assert result.exit_code == 0
    assert "numFilesAdded" in result.output

def test_vacuum(cli_delta_table_with_data, cli_delta_table_path):
  runner = CliRunner()
  result = runner.invoke(vacuum, [cli_delta_table_with_data,
                                  "--retention-hours",
                                  0,
                                  "--force",
                                  "--disable-retention-duration"]
                        )
  assert result.exit_code == 0
  assert "parquet" in result.output


def test_create_checkpoint(cli_delta_table_with_data, cli_delta_table_path):
    runner = CliRunner()
    result = runner.invoke(create_checkpoint, [cli_delta_table_with_data])

    assert result.exit_code == 0
    assert "Checkpoint created successfully" in result.output

def test_table_version(cli_delta_table_with_data):
    runner = CliRunner()
    result = runner.invoke(table_version, [cli_delta_table_with_data])

    assert result.exit_code == 0
    assert result.output.strip().isdigit()


from deltalake import DeltaTable, write_deltalake
import pandas as pd
import logging
import os
import pytest
import boto3

import signal
import requests
import time
import shutil
import subprocess as sp

from click.testing import CliRunner
from deltalake_tools.cli.cli import (
    vacuum,
    compact,
    create_checkpoint,
    table_version
)

from deltalake_tools.models.models import VirtualAddressingStyle

from deltalake_tools.core.core import (
    delta_compact,
    delta_vacuum,
    delta_create_checkpoint,
    table_version as delta_table_version
)


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def start_service(service_name, host, port):
    moto_svr_path = shutil.which("moto_server")
    args = [moto_svr_path, "-H", host, "-p", str(port)]
    process = sp.Popen(
        args, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE
    )  # shell=True
    url = "http://{host}:{port}".format(host=host, port=port)
 
    for i in range(0, 30):
        output = process.poll()
        if output is not None:
            print("moto_server exited status {0}".format(output))
            stdout, stderr = process.communicate()
            print("moto_server stdout: {0}".format(stdout))
            print("moto_server stderr: {0}".format(stderr))
            pytest.fail("Can not start service: {}".format(service_name))
 
        try:
            # we need to bypass the proxies due to monkeypatches
            requests.get(url, timeout=5)
            break
        except requests.exceptions.ConnectionError:
            time.sleep(0.5)
    else:
        stop_process(process)  # pytest.fail doesn't call stop_process
        pytest.fail("Can not start service: {}".format(service_name))
 
    return process
 

def stop_process(process):
    try:
        process.send_signal(signal.SIGTERM)
        process.communicate(timeout=20)
    except sp.TimeoutExpired:
        process.kill()
        time.sleep(3)
        outs, errors = process.communicate(timeout=20)
        exit_code = process.returncode
        msg = "Child process finished {} not in clean way: {} {}".format(
            exit_code, outs, errors
        )
        raise RuntimeError(msg)
 
@pytest.fixture(scope="session")
def s3_delta_table_path(s3_bucket, s3_details):
    delta_table_path = f"s3://{s3_bucket}/cli-delta-table"

    storage_options_result = s3_details.to_s3_config()
    assert storage_options_result.is_ok()

    storage_options = storage_options_result.unwrap()
    # logger.error(f"{storage_options=}")

    for i in range(11):
        data = pd.DataFrame({
            "id": [i+2, i+3],
            "name": [f"{str(i)}_name", f"{str(i)}_other_name"]
        })

        write_deltalake(delta_table_path,
                        data,
                        mode="append",
                        storage_options=storage_options
                    )

    # write_deltalake(delta_table_path, data, mode="append", storage_options=storage_options)

    assert delta_table_path
    dt = DeltaTable(delta_table_path, storage_options=storage_options)

    assert dt.version() == 10

    yield delta_table_path


@pytest.fixture(scope="session")
def s3_server():
    host = "localhost"
    port = 5002
    url = "http://{host}:{port}".format(host=host, port=port)
    process = start_service("s3", host, port)
    logger.info(f"{url=}")
    yield url
    requests.post(f"http://{host}:{port}/moto-api/reset")
    stop_process(process)



@pytest.fixture(scope='session')
def s3_bucket(s3_server):

    s3_client = boto3.client('s3', 
                    aws_access_key_id="test-access-key", 
                    aws_secret_access_key="test-secret-key", 
                    region_name="us-east-1",
                    endpoint_url=s3_server
                )
    bucket_name = 'test-bucket'
    s3_client.create_bucket(Bucket=bucket_name)
  
    yield bucket_name

def test_s3_delta_table(s3_bucket, s3_details, s3_delta_table_path):
    storage_options_result = s3_details.to_s3_config()
    assert storage_options_result.is_ok()
    storage_options = storage_options_result.unwrap()
    dt = DeltaTable(s3_delta_table_path, storage_options=storage_options)

    assert dt.version() == 10

def test_s3_compact(s3_bucket, s3_details, s3_delta_table_path):
    dt = DeltaTable(s3_delta_table_path, storage_options=s3_details.to_s3_config().unwrap())
    assert dt.version() == 10
    result = delta_compact(s3_delta_table_path, s3_details)

    assert result.is_ok()
    assert "numFilesAdded" in result.unwrap()

    dt = DeltaTable(s3_delta_table_path, storage_options=s3_details.to_s3_config().unwrap())
    assert dt.version() == 11

def test_s3_vacuum(s3_bucket, s3_details, s3_delta_table_path):
    dt = DeltaTable(s3_delta_table_path, storage_options=s3_details.to_s3_config().unwrap())
    assert dt.version() == 11
    result = delta_vacuum(s3_delta_table_path,
                            s3_details,
                            retention_hours=0,
                            enforce_retention_duration=False,
                            dry_run=False)

    assert result.is_ok()
    assert any(["parquet" in f for f in result.unwrap()])
    assert len(result.unwrap()) == 11

    dt = DeltaTable(s3_delta_table_path, storage_options=s3_details.to_s3_config().unwrap())
    assert dt.version() == 12

def test_s3_create_checkpoint(s3_bucket, s3_details, s3_delta_table_path):
    dt = DeltaTable(s3_delta_table_path, storage_options=s3_details.to_s3_config().unwrap())
    assert dt.version() == 12
    result = delta_create_checkpoint(s3_delta_table_path, s3_details)

    assert result.is_ok()

def test_s3_table_version(s3_bucket, s3_details, s3_delta_table_path):
    dt = DeltaTable(s3_delta_table_path, storage_options=s3_details.to_s3_config().unwrap())
    assert dt.version() == 12
    result = delta_table_version(s3_delta_table_path, s3_details)
    assert isinstance(result.unwrap(), int)

    assert result.is_ok()
    assert result.unwrap() == 12

    s3_client = boto3.client('s3', 
                aws_access_key_id="test-access-key", 
                aws_secret_access_key="test-secret-key", 
                region_name="us-east-1",
                endpoint_url="http://localhost:5002"
            )
    bucket_name = 'test-bucket'

    files = s3_client.list_objects_v2(Bucket=bucket_name, Prefix="cli-delta-table/_delta_log/_last_checkpoint")
    # logger.error(f"{files=}")
    assert files["KeyCount"] == 1


@pytest.mark.cli
def test_cli_table_version(s3_bucket, s3_details, s3_delta_table_path):

    args = [
        s3_delta_table_path,
        "--bucket", s3_bucket,
        "--access-key-id", s3_details.hmac_keys.access_key_id,
        "--secret-access-key", s3_details.hmac_keys.secret_access_key,
        "--region", s3_details.region,
        "--endpoint-host", s3_details.endpoint_host,
        "--port", s3_details.port,
        "--scheme", s3_details.scheme.value,
    ]

    if s3_details.allow_unsafe_https:
        args.append("--allow-unsafe-https")
    
    if s3_details.virtual_addressing_style == VirtualAddressingStyle.Path:
        args.append("--path-addressing-style")

    runner = CliRunner()
    result = runner.invoke(table_version, args)
    assert result.output.strip().isdigit()
    assert int(result.output.strip()) == 12
    assert result.exit_code == 0
